// constraints.c -- Larry Medwin 10/27/92
// This is based on the constraint system of Abelson and Sussman, p 230.

#include "constraints.h"
// #include "icalc.h"

// for class SymbolList
implementPtrList( SymList, Sym)

// Global symbol table for lex
SymbolList* Symtab;

//////////////////////
// class Connector
//////////////////////
Connector::Connector() : top_(0), bottom_(0), setter_(0) { }

float Connector::get_value() { return value_; }

boolean Connector::has_value() { return (setter_ != 0); }

void Connector::set_value( float value, Constraint* informant)
{
    assert (informant != 0);
    assert (informant == top_ || informant == bottom_);

    // Already set by some other informant?
    if (has_value() && informant != setter_)
        return;

    // Save new value
    value_ = value;
    setter_ = informant;

    // Pass "process_new_value" message to other constraint
    if (informant == top_ && bottom_ != 0) {
        bottom_->process_new_value();
    }
    else if (informant == bottom_ && top_ != 0) {
        top_->process_new_value();
    }
    return;
}

void Connector::forget_value( Constraint* informant)
{
    assert (informant != 0);
    assert (informant == top_ || informant == bottom_);

    // Has no value; nothing to do
    if (!has_value())
        return;

    // Was I set by some other informant?
    if (informant != setter_)
        return;

    // Pass "forget_new_value" message to other constraint
    if (informant == top_ && bottom_ != 0) {
        setter_ = 0;
        bottom_->process_forget_value(this);
    }
    else if (informant == bottom_ && top_ != 0) {
        setter_ = 0;
        top_->process_forget_value(this);
    }
    return;
}

void Connector::connect( Constraint* new_constraint)
{
    assert (new_constraint != 0);
    assert (top_ == 0 || bottom_ == 0);

    if (top_ == 0) {
        top_ = new_constraint;
    }
    else if (bottom_ == 0) {
        bottom_ = new_constraint;
        assert (top_ != bottom_);
    }
    return;
}

// Get ready to delete a constraint
// Be careful to unconstrain network if old constraint was forcing
void Connector::disconnect( Constraint* old_constraint)
{
    assert (old_constraint != 0);
    assert (top_ == old_constraint || bottom_ == old_constraint);

    if (top_ == old_constraint) {
        if(has_value() && top_ == setter_) {
            if(bottom_ != 0) {
                bottom_->process_forget_value(this);
            }
        }
        top_ = 0;
    }
    else if (bottom_ == old_constraint) {
        if(has_value() && bottom_ == setter_) {
            if(top_ != 0) {
                top_->process_forget_value(this);
            }
        }
        bottom_ = 0;
    }
    return;
}

// Get ready to be deleted
void Connector::replace( Connector* new_one)
{
    assert (new_one != 0);

    // Tell the constraint to swap connectors
    if( top_) {
        top_->replace_connector(this, new_one);
    }
    else if( bottom_) {
        bottom_->replace_connector(this, new_one);
    }
    return;
}

// Get ready to be deleted
void Connector::process_delete(Constraint* terminator)
{
    assert (terminator != 0);

    // find other end
    if (terminator == top_) {
        bottom_->process_delete(this);
        delete bottom_;
    }
    else if (terminator == bottom_) {
        top_->process_delete(this);
        delete top_;
    }
    return;
}


//////////////////////
// class Binop 
//////////////////////
Binop::Binop( Connector* con1, Connector* con2, Connector* con_result) :
    op1 (con1), op2 (con2), result (con_result) {

    // Attach to connectors
    op1->connect(this);
    op2->connect(this);
    result->connect(this);
}

void Binop::process_new_value()
{
    if( op1->has_value() && op2->has_value()) {
        result->set_value( forward( op1->get_value(), op2->get_value()), this);
    }
    else if( op1->has_value() && result->has_value()) {
        op2->set_value( inverse( result->get_value(), op1->get_value()), this);
    }
    else if( op2->has_value() && result->has_value()) {
        op1->set_value( inverse( result->get_value(), op2->get_value()), this);
    }
    // else Binop is underconstrained, don't pass along message
    return;
}

void Binop::process_forget_value( Connector* retractor)
{
    // Pass "forget_value" message to all attached connectors
    //   except for retractor
    //   (only clears those previously set by this Binop)
    if (op1 != retractor) {
        op1->forget_value(this);
    }
    if (op2 != retractor) {
        op2->forget_value(this);
    }
    if (result != retractor) {
        result->forget_value(this);
    }

    // Propagate new state
    process_new_value();

    return;
}

void Binop::process_delete(Connector* terminator)
{
    if (terminator != op1) {
        op1->process_delete(this);
        delete op1;
    }
    if (terminator != op2) {
        op2->process_delete(this);
        delete op2;
    }
    if (terminator != result) {
        result->process_delete(this);
        delete result;
    }
    return;
}

void Binop::replace_connector(Connector* old, Connector* new_one)
{
    assert (old == op1 || old == op2 || old == result);

    // Swap the specified Connectors
    if( old == op1) {
        op1 = new_one;
    }
    else if( old == op2) {
        op2 = new_one;
    }
    else if( old == result) {
        result = new_one;
    }

    // Tell the new Connector he's connected
    new_one->connect(this);

    // Propagate any new constraints
    process_new_value();

    return;
}

////////////////////////////////////
// class Multiplier x * y = z
////////////////////////////////////
Multiplier::Multiplier( Connector* m1, Connector* m2, Connector* product) :
    Binop( m1, m2, product)
{
    // Propagate state, if any
    process_new_value();
}

float Multiplier::forward( float x, float y) {
    return (x * y);
}

float Multiplier::inverse( float z, float x) {
    if (( x < .0001 ) && (x > - .0001)) {
        cout << "Divide by zero error in multiplier.\n";
        cout.flush();
        return (1e20);      //  return HUGE_VAL;
    }
    else
        return (z / x);
}

////////////////////////////////////
// class Adder x + y = z
////////////////////////////////////
Adder::Adder( Connector* a1, Connector* a2, Connector* sum) :
    Binop( a1, a2, sum)
{
    // Propagate state, if any
    process_new_value();
}

float Adder::forward( float x, float y)
{
    return (x + y);
}

float Adder::inverse( float z, float x) {
    return (z - x);
}

////////////////////////////////////
// class Constant
////////////////////////////////////

// Constant state
boolean Constant::has_value() { return True;};
float Constant::value() { return value_;}

void Constant::process_new_value()
{
    assert(True);
    return;
}

void Constant::process_forget_value(Connector* retractor)
{
    // Should never call this guy
    assert(True);
    return;
}

// Set up links in both directions
void Constant::connect( Connector* con) {
    con_ = con;
    con->connect(this);

    // Propagate value of Constant
    con->set_value( value(), this);

    return;
}

void Constant::replace_connector(Connector* old, Connector* new_one)
{
    assert (con_ == old);

    // Replace connector
    con_ = new_one;

    // Tell the new Connector he's connected
    new_one->connect(this);

    // Propagate any new constraints
    process_new_value();

    return;
}


////////////////////////////////////
// class Sym
////////////////////////////////////

int Sym::state() { return state_; };

void Sym::state( int new_state)
{
    int old_state = state();

    // Make the change
    state_ = new_state;

    // Indicate state changed, if it has
    if (state() != old_state) {
        show_state();
    }
    return;
}

float Sym::value() { return value_;}

void Sym::value( float val )
{
    value_ = val;
    return;
}

void Sym::set_value( float val )
{
    // Shouldn't try to set this value if it is being forced
    assert( state() != PASSIVE);

    // Remove last constraint
    con_->forget_value(this);

    // Update state and value
    value( val);
    state( ACTIVE_WITH_VALUE );

    // Have connector pick up new value
    con_->set_value( value_, this);

    return;
}

void Sym::unset_value()
{
    // Can only unset if Sym is ACTIVE_*
    assert( state() == ACTIVE_WITH_VALUE || state() == ACTIVE_NO_VALUE);

    // Update stae
    state( UNCONSTRAINED);

    // Unconstrain the connector
    con_->forget_value( this);

    return;
}

// Is Sym forcing its value or does its connector have value?
boolean Sym::has_value()
{
    return (state() == ACTIVE_WITH_VALUE || con_->has_value());
}

// Symbol stuff
const char* Sym::name()
{
    return name_.string();
};

boolean Sym::match_name(char* s) {
    return !::strncmp( name_.string(), s, name_.length());
}

// Sym gets new value from connector
void Sym::process_new_value()
{
    if (state() == UNCONSTRAINED || state() == PASSIVE) {

        // Does the connector have value?
        if(con_->has_value()) {

            // Sym has or is getting a value
            state( PASSIVE);

            // Get the value from the connector
            value( con_->get_value());

            // Tell the user
            show_val();
        }
    }
    return;
}

// Connector removes constraint; tell the other Connectors
void Sym::process_forget_value(Connector* retractor)
{
    assert( retractor == con_);

    if (state() == PASSIVE) {

        // State has changed
        state(UNCONSTRAINED);
        show_state();

        // Tell user
        show_no_val();
    }
    return;
}

// Set up links in both directions
void Sym::connect( Connector* con)
{
    con_ = con;
    con_->connect(this);

    // Is the Sym forcing its value?
    if (state() == ACTIVE_WITH_VALUE) {
        con->set_value( value(), this);
    }
    return;
}

void Sym::replace_connector(Connector* old, Connector* new_one)
{
    assert (con_ == old);

    // Replace it
    con_ = new_one;

    // Tell the new Connector he's connected
    new_one->connect(this);

    // Propagate any new constraints
    process_new_value();

    return;
}

void Sym::show_val()
{
    cout << name() << " is set to " << value() << "\n";
    cout.flush();
    return;
}

void Sym::show_no_val()
{
    cout << name() << " lost its value.\n";
    cout.flush();
    return;
}

///////////////////////////////
// class SymbolList
///////////////////////////////
SymbolList::~SymbolList()
{
Sym* symbol;

    // Find the first Sym; use as root for recursive delete
    ListUpdater(SymList) i(*this);

    if( ! i.more())
        return;

    symbol = i.cur();

    // Recursively delete constraint network
    if( symbol != 0) {
        symbol->con_->process_delete(i.cur());
    }

    // Delete "root"
    delete symbol;

    // Remove all Syms from list
    remove_all();
}

// Find a variable in the Symbol table, or return 0
Sym* SymbolList::sym_find(char* name)
{
    // Is name in Symtab?
    for (ListItr(SymList) i(*this); i.more(); i.next()) {

        Sym* symbol = i.cur();

        // If it is, return a pointer to the element
        if( symbol->match_name(name)) {
            return symbol;
        }
    }

    // Otherwise
    return 0;
}

// Lookup, and insert new variables in symbol table, if necessary.
Sym* SymbolList::sym_lookup(char* name)
{
    Sym* symbol;

    if(( symbol = sym_find(name)) != 0)
        return symbol;

    // Didn't find it, so save new Sym into the symbol table
    Sym* newsym = new Sym(name);
    append( newsym);

    return newsym;
}
