// constraints.h -- Larry Medwin 10/27/92
// This is based on the constraint system of Abelson and Sussman, p 230.
// There should be no UI dependencies in this file (although iv Lists
//   and Strings are ok)

#ifndef __CONSTRAINTS_H__
#define __CONSTRAINTS_H__

#include <OS/list.h>
#include <OS/string.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>

#include <assert.h>

#define False 0
#define True !False

extern class Connector;
extern class Constraint;
extern class    Binop;
extern class        Multiplier;
extern class        Adder;
extern class    Constant;
extern class    Sym;
extern class        SymList;
extern class            SymbolList;

extern SymbolList* Symtab;

// For the parser
extern int Phase;           // Only add symbols while parsing equations
#define INIT 0
#define PARSE 1
#define COMMAND 2

// The connector handles the constraint protocols
//   to implement the propagation of constraints
//   in the constraint network.
class Connector {
    float value_;
    Constraint* top_;       // Constraint at one end
    Constraint* bottom_;    // ... at the other end
    Constraint* setter_;    // Whoever forces this connector's value, or 0

  public:
    Connector();
    ~Connector() {};

    // Accessing value
    void set_value( float, Constraint*);
    void forget_value( Constraint*);
    float get_value();
    boolean has_value();

    // Connection to network
    void connect(Constraint*);
    void disconnect(Constraint*);
    void replace(Connector*);
    void process_delete(Constraint*);
};

// Use this base class to allow the Constraint* members to point to any 
// subclass of Constraint
class Constraint {
  public:
    // The following are recursive, depth first tree-walks,
    //   using informant or retractor as roots.
    virtual void process_new_value() = 0;
    virtual void process_forget_value( Connector*) = 0;
    virtual void replace_connector(Connector*, Connector*) = 0;
    virtual void process_delete(Connector*) = 0;
};

class Binop : public Constraint {
  protected:
    Connector* op1;
    Connector* op2;
    Connector* result;
  public:
    Binop( Connector*, Connector*, Connector*);
    virtual ~Binop() {};

    // Propagation of value
    virtual void process_new_value();
    virtual void process_forget_value( Connector*);

    // Constraint network stuff
    virtual void replace_connector(Connector*, Connector*);
    virtual void process_delete(Connector*);

    // The operation
    virtual float forward( float, float) = 0;
    virtual float inverse( float, float) = 0;
};

class Multiplier : public Binop {
  public:
    Multiplier( Connector*, Connector*, Connector*);
    ~Multiplier() {};

    // x * y = z
    float forward( float, float);
    float inverse( float, float);
};

class Adder : public Binop {
  public:
    Adder( Connector*, Connector*, Connector*);
    ~Adder(){};

    // x + y = z
    float forward( float, float);
    float inverse( float, float);
};

class Constant : public Constraint {
  public:
    Constant::Constant( float value) : value_(value) {};
    ~Constant() {};

    // Constant state
    boolean has_value();
    float value();

    // Propagation of constraints
    virtual void process_new_value();
    virtual void process_forget_value(Connector*);

    // Constraint network stuff
    void connect(Connector*);
    virtual void replace_connector(Connector*, Connector*);
    void process_delete(Connector*) {};

  private:
    float value_;
    Connector* con_;        // Network connection
};

// Sym writes to cout; yacc sets and unsets its value
class Sym : public Constraint {
  public:
    Sym::Sym(const char* name) : state_(UNCONSTRAINED), name_(name) {};
    virtual ~Sym() {};

    // Sym state
    enum { UNCONSTRAINED, PASSIVE, ACTIVE_NO_VALUE, ACTIVE_WITH_VALUE };
    int state();

    // Sym value
    float value();
    virtual void set_value( float);
    virtual void unset_value();
    boolean has_value();

    // Symbol stuff
    const char* name();
    boolean match_name(char*);

    // Propagation of constraints
    virtual void process_new_value();
    virtual void process_forget_value(Connector*);

    // Constraint network stuff
    void connect(Connector*);
    virtual void replace_connector(Connector*, Connector*);
    void process_delete(Connector*) {};

    // Network connection
    Connector* con_;

  protected:
    void value( float val);
    void state(int);

    // User Interface stuff
    virtual void show_val();
    virtual void show_no_val();
    virtual void show_state() {};

    int state_;
    float value_;

  private:
    CopyString name_;
};

// class SymList is an InterViews PtrList of Syms
declarePtrList( SymList, Sym)

// SymbolList class handles the symbol table
class SymbolList : public SymList {
  public:
    SymbolList() {};
    virtual ~SymbolList();          // Clean up network and symtab

    // Access to Symbol table
    Sym* sym_find(char *);          // Lookup only
    Sym* sym_lookup(char *);        // Lookup and insert if not found
};

#endif // __CONSTRAINTS_H__
