/*
  constraint calculator grammar - LM - 2/13/93

  Building the constrained expression tree
  1. Each expression, CONSTANT, or VARIABLE creates a Connector
  and returns it as its value.
  2. See Schreiner and Freidman, p 103 to see how to declare type of $$.
  3. See Schreiner and Freidman, p 50 for desk calculator.
  4. See Schreiner and Freidman, p 54 for parsing floats.
*/

%{
#include "constraints.h"
#include <stream.h>

int Phase;                  // either PARSE or COMMAND
%}

/* Define types passed on value stack */
%union {
    Sym* y_sym;             // Return reference to symbol table element
    char* y_char;
    Connector* y_con;       // Return Connector* from new
    float y_float;
    }

/* Terminal symbols */
%token <y_sym> VARIABLE
%token <y_float> CONSTANT
%token '='
%token '+'
%token '-'
%token '*'
%token '/'
%token '('
%token ')'
%token SET
%token UNSET
%token EQUATION

/* Typed non-terminal symbols */
%type <y_con> expression

/* Precedence */
%left '+' '-'
%left '*' '/'

%%

session
    : sequences

sequences
    : sequence
    | sequence sequences

sequence
    : EQUATION equation commands

equation
    : expression '=' expression
        {
            $1->replace($3);        // Only need one Connector
            delete $1;              //   to attach the two Constraints

            Phase = COMMAND;        // Done with PARSE Phase
        }

expression
    : expression '+' expression     /* $1 + $3 = $$ */
        {
            $$ = new Connector;
            new Adder( $1, $3, $$);
        }
    | expression '-' expression     /* $1 - $3 = $$, so $3 + $$ = $1 */
        {
            $$ = new Connector;
            new Adder( $3, $$, $1);
        }
    | expression '*' expression     /* $1 * $3 = $$ */
        {
            $$ = new Connector;
            new Multiplier( $1, $3, $$);
        }
    | expression '/' expression     /* $1 / $3 = $$, so $3 * $$ = $1 */
        {
            $$ = new Connector;
            new Adder( $3, $$, $1);
        }
    | CONSTANT
        {
            $$ = new Connector;
            Constant* tmp = new Constant( $1);
            tmp->connect($$);
        }
    | VARIABLE
        {
            $$ = new Connector;
            $1->connect($$);
        }
    | '(' expression ')'
        {
            $$ = $2;
        }

/*  Commands control and query a Sym's value and state. */
commands
    : command
    | command commands

command
    : /* empty */
    | SET VARIABLE CONSTANT
        {
            if ($2 == 0) {
                cout << "No such variable.\n";
                cout.flush();
            }
            else if ($2->has_value() && $2->state() == Sym::PASSIVE) {
                cout << $2->name() << " already has a value.\n";
                cout.flush();
            }
            else {
                $2->set_value( $3);
            }
        }
    | UNSET VARIABLE
        {
            if ($2 == 0) {
                cout << "No such variable.\n";
                cout.flush();
            }
            /* Is Sym active? */
            else if ($2->state() == Sym::ACTIVE_WITH_VALUE
             || $2->state() == Sym::ACTIVE_NO_VALUE) {

                /* ok to unset */
                $2->unset_value();
            }
            else {
                cout << $2->name() << " is not forcing a value.\n";
                cout.flush();
            }
        }
    | VARIABLE
        {
            if ($1 == 0) {
                cout << "No such variable.\n";
                cout.flush();
            }
            else if($1->has_value()) {
                cout << $1->name() << " has value " << $1->value() << "\n";
                cout.flush();
            }
            else {
                cout << $1->name() << " has no value.\n";
                cout.flush();
            }
        }


%%

#include <stdio.h>

void yyerror(char* s) { printf("%s\n", s); }

int yywrap() { return 1; }
