// icalc.c -- Larry Medwin 4/5/93
// InterViews user interface for the constraint system

#include <stream.h>
#include "icalc.h"

// Set by App::App(), read by IVSym::IVSym()
Style* IVSym::sym_style;

// Global I/O handlers for lex
App* Myapp;


implementFieldEditorCallback(IVSym)
implementActionCallback(IVSym)

////////////////////////////////////
// class IVSym
////////////////////////////////////

IVSym::IVSym( Sym* source_sym) : Sym(source_sym->name())
{
    // Copy state_ and value_
    state_ = source_sym->state();
    value_ = source_sym->value();

    IVSym_init();
}

void IVSym::IVSym_init()
{
    WidgetKit& kit = *WidgetKit::instance();
    const LayoutKit& layout = *LayoutKit::instance();

    // Init field editors (one for editing, one for uneditable display)
    kit.begin_style("ed");
    ed_ = DialogKit::instance()->field_editor(
        "       ", sym_style,
        new FieldEditorCallback(IVSym)(
            this, &IVSym::accept_editor, &IVSym::cancel_editor)
    );
    kit.end_style();

    kit.begin_style("uned");
    uned_ = DialogKit::instance()->field_editor(
        "       ", sym_style,
        new FieldEditorCallback(IVSym)(
            this, &IVSym::cancel_editor, &IVSym::cancel_editor)
    );
    kit.end_style();

    // Put ed and uned in deck
    //   arrange the glyphs in the deck
    //   in the same order as enum state_ values,
    //   so that flip_to can index the proper glyph by state
    deck_ = layout.deck( uned_, uned_, ed_, ed_);
    deck_->flip_to( state());

    patch_ = new Patch( deck_);

    // The Pushbutton
    pbutton_ = kit.push_button("Force value",
        new ActionCallback(IVSym)(this, &IVSym::click_me));

    // Put all this stuff in an hbox
    body(
        layout.vbox(
            layout.vglue(),
            layout.hbox(
                layout.hglue(), kit.label(name()),  // Sym name
//              layout.hglue(), kit.label("  "),    // spaces
                layout.hglue(), patch_,             // deck
                layout.hglue(), pbutton_,
                layout.hglue()
            ),
            layout.vglue()
        ) 
    );

    // Start these guys up
    ed_->field("");
    uned_->field("");

    return;
};

// Implement state transition
void IVSym::click_me()
{
    switch(state()) {
        case UNCONSTRAINED:
            state(ACTIVE_NO_VALUE);
            break;
        case PASSIVE:       // Can't make transition if Sym is constrained
            break;
        case ACTIVE_WITH_VALUE:
            state(UNCONSTRAINED);
            con_->forget_value( this );
            break;
        case ACTIVE_NO_VALUE:
            state(UNCONSTRAINED);
            break;
        default:
            assert(True);
    }
    // see /home/iv/src/examples/zoomer/main.c
    patch_->redraw();

    return;
}

void IVSym::accept_editor(FieldEditor* ed)
{
    float ed_val;

    // Shouldn't try to set this value if it is being forced
    assert( state() != PASSIVE);


    // Get value from field editor
    const String& valueStr = *ed->text();

    // Illegal number?
    if (!valueStr.convert(ed_val)) {
        cancel_editor(ed);
    }
    else {
        // Remove last constraint
        con_->forget_value(this);

        // Update state and propagate value
        state(ACTIVE_WITH_VALUE);
        set_value(ed_val);
    }
    return;
};

void IVSym::cancel_editor(FieldEditor* ed)
{
    // Restore old value
    switch(state()) {
        case PASSIVE:
        case ACTIVE_WITH_VALUE:
            show_val();
            break;
        case UNCONSTRAINED:
        case ACTIVE_NO_VALUE:
            show_no_val();
            break;
        default:
            assert(True);
    }
    return;
};

void IVSym::show_val()
{
char tmp[100];

    assert( state() != ACTIVE_NO_VALUE);
    assert( state() != UNCONSTRAINED);

    switch(state()) {
        case ACTIVE_WITH_VALUE:
            sprintf( tmp, "%6.2f", value());
            ed_->field(tmp);
            break;
        case PASSIVE:
            sprintf( tmp, "%6.2f", con_->get_value());
            uned_->field(tmp);
            break;
        default:
            assert(True);
    }
    return;
}

void IVSym::show_no_val()
{
    assert( state() != ACTIVE_WITH_VALUE);
    assert( state() != PASSIVE);

    switch(state()) {
        case ACTIVE_NO_VALUE:
            ed_->field("");
            break;
        case UNCONSTRAINED:
            uned_->field("");
            break;
        default:
            assert(True);
    }
    return;
}

void IVSym::show_state()
{
    deck_->flip_to(state());
    patch_->redraw();
    return;
}

///////////////////////////////
// class Lexio
///////////////////////////////
Lexio::Lexio( String equation) {

    // Copy into lex input string
    // Format: "equation " %s '\0'
    equation_ = new char[equation.length()+strlen("equation ")+1];
    sprintf( equation_, "equation %s", equation.string());
    current_ = equation_;
}

Lexio::~Lexio()
{
    delete[] equation_;
}

int Lexio::lex_input()
{
    // Don't step back over start of string...
    if (current_ < equation_) {
        return 0;       // EOF
    }
    // Return char, or 0 at end of string
    else {
        return *(current_++);
    }
}

void Lexio::unput(char c)
{
    // If not before beginning
    if (current_ > equation_)
        // Push back pointer and put in char
        *(--current_) = c;

    return;
}


///////////////////
// class App
///////////////////

implementFieldEditorCallback(App)

App::App(Style* style) : network_(0)
{
    const LayoutKit& layout = *LayoutKit::instance();

    // Save copy of style
    IVSym::sym_style = new Style(style);

    // Init field editor
    eq_editor_ = DialogKit::instance()->field_editor(
        "                    ", IVSym::sym_style,
        new FieldEditorCallback(App)(
            this, &App::accept_editor, &App::cancel_editor
        )
    );

    // Init CurSymArea and its enclosing box
    sym_box_ = layout.vbox();
}

void App::accept_editor(FieldEditor* ed)
{
    IVSym* tmp_sym;
    WidgetKit& kit = *WidgetKit::instance();

    // Get rid of old constraint network
    // if( network_ != 0) {
    //  network_->process_delete((Connector*)0);
    //  delete network_;
    //}

    // Clean up glyphs
    for (GlyphIndex g = sym_box_->count(); g > 0; g--) {
        sym_box_->remove(g-1);
    }

    // Get FieldEditor string to lex
    lex_handler = new Lexio( *ed->text());

    // Create and empty symbol table
    delete Symtab;
    Symtab = new SymbolList;

    // Parse the equation, generate constraint network and symbol table
    Phase = PARSE;
    yyparse();

    // Build up sym_box area
    for (ListUpdater(SymList) i(*Symtab); i.more(); i.next()) {

        Sym* cur_sym = i.cur();

        // "Promote" Sym to IVSym
        tmp_sym = new IVSym( cur_sym);

        // Replace Sym with IVSym in the constraint network
        Connector* tmp_con = cur_sym->con_;
        tmp_con->disconnect(cur_sym);
        tmp_sym->connect(tmp_con);

        // delete cur_sym;

        // Put the IVSym in the sym_box_ area
        sym_box_->append(tmp_sym);
    }

    // Keep a hook into the network so we can clean up later
    // network_ = tmp_sym;

    // UI gives us access to symbols, not the symbol table
    // delete Symtab;

    // Send expose event (How??)

    return;
}

void App::cancel_editor(FieldEditor* ed)
{
    // Restore old equation
    ed->field(equation_);
    return;
};

// Start it off
void App::kick_start(char* s)
{
    // Set up initial string
    eq_editor_->field( s );

    // Get started
    accept_editor( eq_editor_);

    return;
}

//////////////////////
//   main()
//////////////////////
int main(int argc, char** argv)
{
    Session* session = new Session("FieldEditorTest", argc, argv);
    WidgetKit& kit = *WidgetKit::instance();
    const LayoutKit& layout = *LayoutKit::instance();
    Myapp = new App(session->style());

    // Start it off
    Myapp->kick_start("a+b+c+d+e=0");

    // Set up app window
    return session->run_window(
        new ApplicationWindow(
            new Background(
                layout.hbox(
                    layout.hglue(),
                    layout.vbox(
                        layout.vglue(), Myapp->eq_editor_,
                        layout.vglue(), Myapp->sym_box_,
                        layout.vglue()
                    ),
                    layout.hglue()
                    ),
                kit.background()
            )
        )
    );
    exit(0);
}
